//---------------------------------------------------------------------------
// jrdebug_commandline.h
// helper functions for commandline stuff
//---------------------------------------------------------------------------




bool JrDebug::ParseCommandline(char *inprogramdesc,int *argcp, char **argv, char* altcomlinestr, char *precomline, char* defaultcomline, char *indefaultoutputoptions)
	{
	// this helper function takes standard argc and argv and look for debug realted commandline args
	//  if none found, the defaultcommandline will be used
	// return true if user should exit right away (ie we ran tests or listed tests or formatters)
	bool shouldexit=false;
	bool foundcommandlineargs=false;

	// before parsing commandline we want to sweep and delete any formatters that may have been created by static objects
	// shut down all outputting formatters
	RemoveInstantiatedFormatters();

	// save default outputoptions in case we need to create an output later
	if (!strcmp(indefaultoutputoptions,""))
		defaultoutputoptions=jrd_stdstringT(indefaultoutputoptions);

	// save program description
	programdescription=jrd_stdstringT(inprogramdesc);

	// first run precomline and reset foundargs
	if (precomline!=NULL)
		{
		ParseCommandlineArgString(precomline,NULL,&foundcommandlineargs);
		foundcommandlineargs=false;
		}

	// now commandline args
	if (argcp!=0)
		{
		// parse commandline args (this will set foundcommandlineargs)
		int argc=*argcp;
		shouldexit = ParseCommandlineArgcArgv(argc,argv,&foundcommandlineargs);
		*argcp=argc;
		}
	else if (altcomlinestr!=NULL)
		{
		// use alternate commandline string (borland c++ use for example)
		shouldexit = ParseCommandlineArgString(altcomlinestr,NULL,&foundcommandlineargs);
		}

	// if no commandline args are found, run default
	if (!foundcommandlineargs)
		{
		// parse the default commandlinestring
		shouldexit = ParseCommandlineArgString(defaultcomline,inprogramdesc,&foundcommandlineargs);
		// reset commandlineargs since we are using defaults not the real commandline args (might be useful for caller to know)
		foundcommandlineargs=false;
		}

	// return whether to exit
	return shouldexit;
	}



bool JrDebug::ParseCommandlineArgString(const char* commandlinestr,char *arg0,bool *foundcommandlineargsp)
	{
	// parse a string as if it was a string of commandline args
	// return true if user should exit right away (ie we ran tests or listed tests or formatters)
	bool shouldexit;
	char dummyarg0[2] = {0}; //blank char* string

	// fill argc and argv
	int argc=0;
	char *argv[JrDebug_MaxCommandlineLargs];

	if (arg0!=NULL)
		{
		// first arg is exename
		argv[argc++]=arg0;
		}
	else
		{
		// empty string
		argv[argc++]=dummyarg0;
		}

	if (commandlinestr==0)
		return false;
	if (strlen(commandlinestr)==0)
		return false;

	// allocate space for comlinebuffer
	int len=(int)(strlen(commandlinestr));
	char *comlinebuf = new char[len+1];

	// now split it
	bool inquote=false;
	char c;
	int comlinepos=0;
	argv[argc]=&comlinebuf[0];
	for (int count=0;count<=len;++count)
		{
		c=commandlinestr[count];
		if (c=='\\' && false)
			{
			// an escape character says next char is ignored? this is not used by default
			c=commandlinestr[++count];
			comlinebuf[comlinepos++]=c;
			}
		else if ((c==34 && inquote==true) || (inquote==false && c==32) || c==0)
			{
			// end of an arg, cap it with 0, and advance argc and set new pointer, assuming this arg was nonblank

			// ATTN: 02/07/07 - bug found by walkerr (http://www.donationcoder.com/Forums/bb/index.php?topic=7361.0)
			inquote=false;

			if (argv[argc]!=&comlinebuf[comlinepos])
				{
				comlinebuf[comlinepos++]='\0';
				argv[++argc]=&comlinebuf[comlinepos];
				if (argc>JrDebug_MaxCommandlineLargs)
					break;
				}
			}
		else if (c==34 && inquote==false)
			{
			// start into a quote
			inquote=true;
			}
		else if (c==32 && argv[argc]==&comlinebuf[comlinepos])
			{
			// a space at the begining of a arg word is ignored
			}
		else
			{
			// its a character so add it to current arg
			comlinebuf[comlinepos++]=c;
			}
		}

	//	// ATTN: test
	//	std::cout << "Parsed string into "<<argc<<" commandline args:"<<std::endl;
	//	for (int count=0;count<argc;++count)
	//		{
	//		std::cout << count << " " << argv[count] << std::endl;
	//		}

	// now invoke parse
	shouldexit=ParseCommandlineArgcArgv(argc,argv,foundcommandlineargsp);

	// free commandline buf
	delete [] comlinebuf;

	// return whether user should exit right away
	return shouldexit;
	}
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
bool JrDebug::ParseCommandlineArgcArgv(int &argc,char **argv,bool *foundcommandlineargsp)
	{
	// if found then debug logging is enable and the file is used for output (overwrite mode)
	// if not found, and disableifnofile is true, then debugging is disabled
	// return true if user should exit right away (ie we ran tests or listed tests or formatters)
	jrd_stdstringT dummystring,keywordstring,dbostring,optionstring;
	bool bretv;
	bool shouldexit=false;
	bool foundcommandlineargs=false;
	bool debuginfoflag=false;

	// save exe name
	char* argv0=argv[0];

	// -dbv option (version announce)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-dbv",dummystring,"",false,false);
	if (bretv)
		{
		// announce and signal we should exit
		Announce("info",true,argc,argv);
		foundcommandlineargs=true;
		shouldexit=true;
		}

	// -dbh option (help)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-dbh",dummystring,"",false,false);
	if (!bretv)
		bretv=FindAndGrabCommandlineArg(argc,argv,"-dbhelp",dummystring,"",false,false);
	if (bretv)
		{
			// announce and signal we should exit
			Announce("help",true,argc,argv);
			CommandlineParseMessage(jrd_stdstringT("Display Version Info     ->   -dbv"));
			CommandlineParseMessage(jrd_stdstringT("Display This Help        ->   -dbh"));
			CommandlineParseMessage(jrd_stdstringT("Enable Debugging Output  ->   -db"));
			CommandlineParseMessage(jrd_stdstringT("Disable Debugging Output ->   -dbd"));
			CommandlineParseMessage(jrd_stdstringT("Add Output Fiter         ->   -dbf"));
			CommandlineParseMessage(jrd_stdstringT("Add Output Object        ->   -dbo STYLE[:options][;FILENAME][;stdout][;sysout][;info]"));
			CommandlineParseMessage(jrd_stdstringT("List Output Styles       ->   -dbol"));
			CommandlineParseMessage(jrd_stdstringT("Run Tests by Keyword     ->   -dbt KEYWORD"));
			CommandlineParseMessage(jrd_stdstringT("List Registered Tests    ->   -dbtl"));
			foundcommandlineargs=true;
			shouldexit=true;
		}

	// -db option (enable debugging, no other info provided)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-db",dummystring,"",false,false);
	if (bretv)
		{
		// turn on debug loggin
		foundcommandlineargs=true;
		SetEnabled(true);
		}

	// -dbd option (disable debugging, no other info provided)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-dbd",dummystring,"",false,false);
	if (bretv)
		{
		// turn off debug loggin
		foundcommandlineargs=true;
		SetEnabled(false);
		}

	// -dbf option (filter)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-dbf",optionstring,"",true,false);
	if (bretv)
		{
		// set some filters
		foundcommandlineargs=true;
		shouldexit=SetFilterOptions(optionstring);
		}


	// -dbo option (output formatter option) (grab as many as available)
	while (FindAndGrabCommandlineArg(argc,argv,"-dbo",dbostring,"",true,false))
		{
		// set up debug output
		foundcommandlineargs=true;
		jrd_stdstringT stylename,filename,optionstring;
		bool stdoutflag,systemoutflag;
		bretv=ParseOutputOption(dbostring,stylename,optionstring,filename,stdoutflag,systemoutflag,debuginfoflag);
		if (bretv)
			{
			// if they specified no output target, default to console and systemout
			if (filename=="" && stdoutflag==false && systemoutflag==false)
				{
				stdoutflag=true;
				systemoutflag=true;
				}




			// ATTN: new 11/13/10
			// IF the output file has no path, figure out a reasonable one to place the output file into
			//  NOTE: this can be complicated because of vista/win7 which dont want an app to write in its own directory (stupid)
			jrd_stdstringT filename_full = AddSmartPathToOutputFile(filename,argv0);
			if (debuginfoflag)
				{
				CommandlineParseMessage(jrd_stdstringT("INFO: Output file path is: ")+ filename_full);
				}

			// instantiate it
			JrDebugRegisteredOutputFormatter* dbgr=InstantiateFormatter(stylename.c_str(),optionstring.c_str(),filename_full.c_str(),stdoutflag,systemoutflag);
			if (dbgr==NULL)
				{
				CommandlineParseMessage(jrd_stdstringT("-dbo option specified unknown '")+stylename+"' ("+dbostring+") failed; use -dbh for help.");
				shouldexit=true;
				}
			else
				SetEnabled(true);
			}
		else
			{
			// error parsing option
			CommandlineParseMessage(jrd_stdstringT("-dbo option specified incorrectly; should be -dbo STYLE[:options][;FILENAME][;stdout][;debug]; use -dbh for help."));
			shouldexit=true;
			}
		}


	// -dbtl option (list all available tests)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-dbtl",keywordstring,"all",true,false);
	if (bretv)
		{
		// open file and enable
		foundcommandlineargs=true;
		SetEnabled(true);
		// list tests
		ListTests();
		shouldexit=true;
		}

	// -dbol option (list all formatting styles)
	bretv=FindAndGrabCommandlineArg(argc,argv,"-dbol",dummystring,"",false,false);
	if (bretv)
		{
		// open file and enable
		foundcommandlineargs=true;
		SetEnabled(true);
		// list formatters
		ListFormatters();
		shouldexit=true;
		}

	// -dbt option (run tests) (grab as many as available)
	while (FindAndGrabCommandlineArg(argc,argv,"-dbt",keywordstring,"all",true,false))
		{
		// open file and enable
		foundcommandlineargs=true;
		SetEnabled(true);
		// run tests
		RunTests(keywordstring.c_str());
		shouldexit=true;
		}

		if (foundcommandlineargsp!=NULL)
			*foundcommandlineargsp=foundcommandlineargs;

	// return true if user should exit right away (ie we ran tests or listed tests or formatters)
	return shouldexit;
	}
//---------------------------------------------------------------------------












//---------------------------------------------------------------------------
bool JrDebug::SetFilterOptions(jrd_stdstringT optionstring)
{
	// set filter options in optionstring
	// return true on error
	// ATTN: unfinished
	if (optionstring.find("e")!=std::string::npos)
		{
		// show only errors
		SetFilter_OnlyErrors(true);
		}
	if (optionstring.find("b")!=std::string::npos)
		{
		// but always show all blocks
		SetFilter_ShowAllBlocks(true);
		}
	return false;
}
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
bool JrDebug::ParseOutputOption(jrd_stdstringT &indbostring, jrd_stdstringT &stylename, jrd_stdstringT &optionstring, jrd_stdstringT &filename,bool &stdoutflag, bool &systemoutflag, bool &debuginfoflag)
{
	// parse the dbo commandline string into its components
	// returns true if args were properly parsed
	jrd_stdstringT dbostring=indbostring;
	stylename="";
	filename="";
	optionstring="";
	stdoutflag=false;
	systemoutflag=false;
	debuginfoflag=false;

	// parse
	//  -dbo STYLE[:options][;FILENAME][;stdout][;sysout][;info]

	// format style (style[:options])
	jrd_stdstringT stylenamepluso;
	if (!SplitStringLeft(dbostring,optionstring,';'))
		return false;
	// split format style into style and options
	if (!SplitStringLeft(optionstring,stylename,':'))
		return false;

	// the rest of the args are filename and other optional stuff
	jrd_stdstringT anargstring;
	while (SplitStringLeft(dbostring,anargstring,';'))
		{
		if (anargstring=="stdout")
			stdoutflag=true;
		else if (anargstring=="sysout")
			systemoutflag=true;
		else if (anargstring=="info")
			debuginfoflag=true;
		else if (filename=="")
			{
			// filename
			filename=anargstring;
			// expand environment strings?
			#ifdef _WIN32JRD
				char tempoutbuf[2056];
				ExpandEnvironmentStrings(filename.c_str(), tempoutbuf, 2055);
				filename=jrd_stdstringT(tempoutbuf);
				// is this second one needed????
				//PathUnExpandEnvStrings(filename.c_str(), tempoutbuf, 2055);
				//filename=jrd_stdstringT(tempoutbuf);
				// ATTN: smart windows version replacement
				// if run on vista (or win7?) the application will not be able to write to its own directory (STUPID IDIOTIC MS DECISION)
				//  so we should fix this up
			#endif
			}
		else
			{
			// hmm they are trying to specify multiple filenames, which is an error
			CommandlineParseMessage(jrd_stdstringT("Attempt to specify multiple output filenames on commandline (")+indbostring+").");
			return false;
			}
		}

	// success
	return true;
}
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
bool JrDebug::SplitStringLeft(jrd_stdstringT &mainstring,jrd_stdstringT &leftpart,char dividerchar)
{
	// split left part of a string off

	int len = (int)(mainstring.length());
	if (len==0)
		{
		// nothing to grab so return false
		leftpart="";
		return false;
		}

	const char *charp=mainstring.c_str();
	bool inquote=false;
	char c;

	// find 
	for (int count=0;count<len;++count)
		{
		c=charp[count];
		if (c==34 && inquote==false)
			inquote=true;
		else if (c==34 && inquote==true)
			inquote=false;
		else if (c==dividerchar && inquote==false)
			{
			// found it, grab left part
			if (count==0)
				leftpart="";
			else
				leftpart=mainstring.substr(0,count);
			// now grab remainder
			if (count<len-1)
				mainstring=mainstring.substr(count+1,len-(count+1));
			else
				mainstring="";
			return true;
			}
		}

	// not found, so its all left part
	leftpart=mainstring;
	mainstring="";
	return true;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void JrDebug::CommandlineParseMessage(const jrd_stdstringT &messagestring)
{
	// display error dont't rely on debug output
	std::cerr << messagestring << std::endl;
}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
bool JrDebug::FindAndGrabCommandlineArg(int &argc,char **argv,const jrd_stdstringT &flagstring, jrd_stdstringT &extrastring, const jrd_stdstringT &defaultextraval, bool grabextra, bool requiresextra)
	{
	// look for commandline arg, if found strip it and return true
	// an extra string can be extracted as well
	int count;
	int removecount=0;
	extrastring="";
	for (count=1;count<argc;++count)
		{
		if (strcmp(argv[count],flagstring.c_str())==0)
			{
			// got one
			++removecount;
			break;
			}
		}
	if (count>=argc)
		return false;

	// filename
	if (count==argc-1)
		{
		if (requiresextra)
			return false;
		extrastring=defaultextraval;
		}
	else if (argv[count+1][0]=='-')
		{
		if (requiresextra)
			return false;
		extrastring=defaultextraval;		
		}
	else if (grabextra)
		{
		// grab extra string
		extrastring=jrd_stdstringT(argv[count+1]);
		++removecount;
		}

	// adjust argc and argv
	for (;count<argc-removecount;++count)
		argv[count]=argv[count+removecount];
	argc-=removecount;

	return true;
	}
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
void JrDebug::Announce(const char *mesg, bool evenifdisabled,int argc,char **argv)
	{
	// announce JrDebug is running to log
	if (!enabled && !evenifdisabled)
		return;
		
	jrd_stdstringT argstring="";
	if (argc>0)
		{
		for (int count=0;count<argc;++count)
			{
			argstring+=jrd_stdstringT(argv[count]);
			if (count<argc-1)
				argstring+=",";
			}
		}

	if (filtermin!=-99999)
		JrDebug( __FILE__ , __LINE__ ,  __TIMESTAMP__ , __FUNCTION__ , __FUNC__, __func__ , JrdMessage)(JrdNote,0,"JrDebug version %s - %s - (enabled=%d asserts=%d exceptions=%d) filterlevel[%d] - ComlineArgs(%d) [%s]",JrDebugVersionString,mesg,(int)enabled,(int)displaypassingasserts,(int)(exceptionlevel!=JrDebugIgnoreLevel),filtermin,argc,argstring.c_str());
	else
		JrDebug( __FILE__ , __LINE__ ,  __TIMESTAMP__ , __FUNCTION__ , __FUNC__, __func__ , JrdMessage)(JrdNote,0,"JrDebug version %s - %s - (enabled=%d asserts=%d exceptions=%d) - ComlineArgs(%d) [%s]",JrDebugVersionString,mesg,(int)enabled,(int)displaypassingasserts,(int)(exceptionlevel!=JrDebugIgnoreLevel),argc,argstring.c_str());
	}
//---------------------------------------------------------------------------































//---------------------------------------------------------------------------
jrd_stdstringT JrDebug::AddSmartPathToOutputFile(jrd_stdstringT filename, char* argv0)
{
	// add a smart path to local filename
	// ATTN: this is tricky because on vista/win7 we MIGHT not be able to write to app directory
	// ask if this has a path, if so 
	// ATTN: alternatively we could just look for ':' if its windows
	if (strchr(filename.c_str(),'/')!=NULL || strchr(filename.c_str(),'\\')!=NULL || argv0==NULL)
		return filename;

	//std::cout << "$DEBUG - STARTING filename for output: "<<filename << std::endl;
	//std::cout << "$DEBUG - argv0: "<<std::string(argv0) << std::endl;

	// first replace any time date stuff
	jrd_stdstringT outfilename;
	outfilename = JrDebugRegisteredOutputFormatter::ReplaceFileNameArgs(filename);

	// ok lets get the dir of the EXE
	jrd_stdstringT exefullpath = GetExeFullPathDirectory(argv0);
	jrd_stdstringT exedir = ExtractDirectoryFromFile(exefullpath);
	jrd_stdstringT exebasename = ExtractBaseNameFomFile(exefullpath);
	//std::cout << "$DEBUG - directory for exe: "<<exedir << std::endl;
	//std::cout << "$DEBUG - basename for exe: "<<exebasename << std::endl;

	bool flag_canwriteindir = CanWeWriteInDirectory(exedir);
	if (flag_canwriteindir)
		{
		// ok so save in app dir
		outfilename = exedir + DEF_DirectorySeparatorString + filename;
		}
	else
		{
		// ok this is trickier case, we need to write somewhere else
		jrd_stdstringT safedirpath = GetSafeDirectoryForWriting("JrDebug/"+exebasename);
		outfilename = safedirpath + DEF_DirectorySeparatorString + filename;
		}

	//std::cout << "$DEBUG - FINAL filename for output: "<<outfilename << std::endl;

	return outfilename;
}


jrd_stdstringT JrDebug::GetExeFullPathDirectory(char* argv0)
{
	jrd_stdstringT exefullpath;
	#ifdef _WIN32JRD
		char buf[MAX_PATH];
		GetModuleFileName(NULL,buf,MAX_PATH);
		buf[MAX_PATH-1]='\0';
		exefullpath=jrd_stdstringT(buf);
#	else
		exefullpath = GetCurrentWorkingDirectory();
		if (argv0!=NULL && argv0[0]!=0)
			{
			if (argv0[0]=='/')
				exefullpath=argv0;
			else
				exefullpath+=DEF_DirectorySeparatorString+jrd_stdstringT(argv0);
			}
	#endif
	return exefullpath;
}

jrd_stdstringT JrDebug::ExtractDirectoryFromFile(jrd_stdstringT filename)
{
	jrd_stdstringT retstr;

	const char *fbuf = filename.c_str();
	const char *sep1 = strrchr(fbuf,'/');
	const char *sep2 = strrchr(fbuf,'\\');
	if (sep2>sep1 && sep2!=NULL)
		sep1=sep2;
	if (sep1!=NULL)
		{
		int len = sep1-fbuf;
		retstr=jrd_stdstringT(filename.c_str(),len);
		}
	else
		retstr=filename;


	return retstr;
}



jrd_stdstringT JrDebug::ExtractBaseNameFomFile(jrd_stdstringT filename)
{
	jrd_stdstringT retstr;

	const char *fbuf = filename.c_str();
	const char *sep1 = strrchr(fbuf,'/');
	const char *sep2 = strrchr(fbuf,'\\');
	if (sep2>sep1 && sep2!=NULL)
		sep1=sep2;
	if (sep1!=NULL)
		{
		retstr=jrd_stdstringT(sep1+1);
		}
	else
		retstr=filename;

	// now pick off extension
	fbuf = retstr.c_str();
	sep1 = strrchr(fbuf,'.');
	if (sep1!=NULL)
		{
		int len = sep1-fbuf;
		retstr=jrd_stdstringT(retstr.c_str(),len);
		}

	return retstr;
}


bool JrDebug::CanWeWriteInDirectory(jrd_stdstringT dirpath)
{
	//return true if we can write in this directory, otherwise false
	//false can happen on vista/win7

	// ok now while you could argue that its not insanely retarded for MS to forbid an app to write in its home directory
	// but combining this with an invisible shadow directory that gets created to emulate this when your app tries it, hiding the shadow dir where it cant be found
	// and makes it nearly impossible to tell if this is happening or if you ahve access to write wher eyou want
	// the only conclusion is the MS is filled with traitors trying to destroy the company.

	// SO.. we will do something less precise but a reasonable heuristic

	// first, if using non-windows or windows <vista, just say yes
	// ATTN: ideally this should be updated for *nix, which could probably be done with a simple quick file creation test
	//  (which doesnt work on win because of FUCKING morons and their secret invisible shadow emulation)

	// if win vista/7 then see if we are in Program Files path, if so then assume we CANNOT write in it
	#ifndef _WIN32JRD
		return true;
	#endif

	#ifdef _WIN32JRD
		OSVERSIONINFO ovi;
		ovi.dwOSVersionInfoSize = sizeof(ovi);
		GetVersionEx(&ovi);
		// older than vista?
		if (ovi.dwPlatformId != VER_PLATFORM_WIN32_NT || ovi.dwMajorVersion < 6)
			return true;
		// ok we are on vista or win7, are we in program files?
		if (strstr(dirpath.c_str(),"Program Files")==NULL)
			return true;
		// else its unsafe
		return false;
	#endif

	return true;
}


jrd_stdstringT JrDebug::GetCurrentWorkingDirectory()
{
   char* buffer;

   // Get the current working directory: 
	#ifdef _WIN32JRD
		buffer = _getcwd( NULL, 0 );
	#else
   		buffer = getcwd( NULL, 0 );
	#endif

	if (buffer==NULL)
		return "";
	return jrd_stdstringT(buffer);
}


jrd_stdstringT JrDebug::GetSafeDirectoryForWriting(jrd_stdstringT subdirpath)
{
	jrd_stdstringT retpath;

	#ifdef _WIN32JRD
		char buf[MAX_PATH];
		SHGetSpecialFolderPath(0, buf, CSIDL_APPDATA, false) ;
		buf[MAX_PATH-1]='\0';
		retpath = jrd_stdstringT(buf);
	#else
		retpath = "~";
	#endif

	jrd_stdstringT topdir;
	while (SplitOffTopDir(subdirpath,topdir))
		{
		if (topdir=="")
			continue;
		retpath+=DEF_DirectorySeparatorString+topdir;
		// make sir
		mkdir(retpath.c_str());
		}

	return retpath;
}

bool JrDebug::SplitOffTopDir(jrd_stdstringT &path, jrd_stdstringT &topdir)
{
	const char *fbuf = path.c_str();
	const char *sep1 = strchr(fbuf,'/');
	const char *sep2 = strchr(fbuf,'\\');
	if (sep2<sep1 && sep2!=NULL)
		sep1=sep2;
	if (sep1==NULL)
		{
		if (path=="")
			return false;
		topdir=path;
		path="";
		return true;
		}
	//
	topdir=jrd_stdstringT(fbuf,sep1-fbuf);
	path=jrd_stdstringT(sep1+1);
	return true;
}
//---------------------------------------------------------------------------




