//---------------------------------------------------------------------------
// jrdebug_switch.h
//  8/6/04 - 2/23/05
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// header guards
#ifndef INC_jrdebug_switchh
#define INC_jrdebug_switchh
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
// This file defines DEFAULT compilation options which will govern all
//  projects you build using JrDebug.  It is usually safe to leave these
//  all at their defaults.
//
// The symbols are defined here in such a way that they can be overridden
//  by the user in -D project/makefile options; so you should view the
//  settings here (inside double #ifdefs) as *default* settings.
//
// So, you can override any of the symbols defined here inside your make file
//  or project file.
// Alternativel you can use a project-specific jrdebug_switch.h file as follows:
//  Make a *copy* of *this* file (jrdebug_switch.h) inside your application
//   directory.  Then *REMOVE* or *MOVE* this file from the main DebugLibrary
//   directory containing jrdebug.h.  (you may have to adjust your project
//   include directory paths to make sure it will look in your application
//   directory).
//  Now, when jrdebug.h tries to #include this file, it will grab the local
//   copy in your application directory.
//  Then it will be safe to change your application debug compilation options
//   from the local project-specific copy of this file.
//---------------------------------------------------------------------------









//---------------------------------------------------------------------------
// default formatter options, if debugging enabled but no output specified
#define JrDebug_DefaultFormatterOptions "-dbo text;stdout -dbo jrd;sysout"
//---------------------------------------------------------------------------



//---------------------------------------------------------------------------
// CONTROLLING THE COMPILATION OF DEBUGGING CODE - DEFAULTS:
// NOTE: Do *not* try to set the defines above from your own .cpp files
//       (see note at top for more info).
// We put all these inside a check for already defined symbols, so you can
//  treat these as default values that can be overridden.

// 1a. Obey Debug flag compiling-in of debug code?
#ifndef JrDebugDIRECTIVE_DontObeyNDEBUG
 #ifndef JrDebugDIRECTIVE_ObeyNDEBUG
  //
  // Usually you want to automatically enable logging when you build
  //  in Debug mode, and not compile-in logging when you build in release mode
  //  (signified when symbol NDEBUG is defined); this also checks for the symbols
  //  DEBUG and __DEBUG__ to indicate a debug compilation switch.
  #define JrDebugDIRECTIVE_ObeyNDEBUG
 #endif
#endif

// 1b. Use JrDebugDIRECTIVE_ObeyNDEBUG to set compilation
#ifndef JrDebugDIRECTIVE_DontCompileDebugging
 #ifndef JrDebugDIRECTIVE_CompileDebugging
  #ifdef JrDebugDIRECTIVE_ObeyNDEBUG
   #ifdef NDEBUG
    #define JrDebugDIRECTIVE_DontCompileDebugging
   #endif
   #ifndef NDEBUG
    #ifndef DEBUG
     #ifndef _DEBUG
      #ifndef __DEBUG__
       #define JrDebugDIRECTIVE_DontCompileDebugging
      #endif
     #endif
    #endif
   #endif
  #endif
 #endif
#endif

// 2. Enable compiling-in of debug code?
#ifndef JrDebugDIRECTIVE_ObeyNDEBUG
 #ifndef JrDebugDIRECTIVE_DontCompileDebugging
  #ifndef JrDebugDIRECTIVE_CompileDebugging
   // Alternatively if you want to always disable all debugging and eliminate all runtime
   //  overhead, unless overriding with -D define, just uncomment this line to
   //  prevent debugging code from being compiled in:
   // #define JrDebugDIRECTIVE_DontCompileDebugging
  #endif
 #endif
#endif

// 3. Start up disabled or enabled? (easily overridden by ParseCommandline func)
#ifndef JrDebugDIRECTIVE_StartupEnabled
 #ifndef JrDebugDIRECTIVE_StartupDisabled
  // If you want to compile in the ability to turn on debugging at runtime, but
  //  turn off debugging at startup, you can define the following symbol (note
  //  it has no effect if debugging is compiled out):
  // #define JrDebugDIRECTIVE_StartupDisabled
 #endif
#endif

// 4. Use minimal commandline arg parsing warning when compiled-out?
#ifndef JrDebugDIRECTIVE_CompiledOutCommandlineParseOn
 #ifndef JrDebugDIRECTIVE_CompiledOutCommandlineParseOff
  // If this symbol is not set then a minimal extra function will be compiled-in
  //  even when debugging is set to not compile-in, which will warn any
  //  user if they try to invoke a debug commandline flag.
  #define JrDebugDIRECTIVE_CompiledOutCommandlineParseOff
 #endif
#endif
//---------------------------------------------------------------------------





//---------------------------------------------------------------------------
// NAMESPACES:
//
#ifndef JrDebugDIRECTIVE_UseNameSpaces
 #ifndef JrDebugDIRECTIVE_DontUseNameSpaces
  // Should entire JrDebug functions+classes be inside the 'jrdebug' namespace?
  //#define JrDebugDIRECTIVE_UseNameSpaces
 #endif
#endif

#ifndef JrDebugDIRECTIVE_UseNameSpacesAutoUsing
 #ifndef JrDebugDIRECTIVE_UseNameSpacesAutoUsing
  // Should we automatically perform a "using namespace jrdbug;"
  //  this only has an effect if above symbol is also defined
  //#define JrDebugDIRECTIVE_UseNameSpacesAutoUsing
 #endif
#endif
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// TAKING OVER BUILT IN MACROS:
// 
#ifndef JrDebugDIRECTIVE_TakeoverAssertTrace
 #ifndef JrDebugDIRECTIVE_DontTakeoverAssertTrace
  // Do you want to take over the calls to TRACE, ASSERT, VERIFY if debugging is enabled?
  //  in general i think this is probably not the best idea, but can be useful for working with legacy code you dont want to modify
  //  #define JrDebugDIRECTIVE_TakeoverAssertTrace
 #endif
#endif
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// SOME sNprintf FUNCTIONS NOT AVAILABLE IN SOME COMPILERS:
//
#ifndef JrDebugDIRECTIVE_DontTryToUseSnprintf
 #ifndef JrDebugDIRECTIVE_TryToUseSnprintf
  // We try to use a version of printf (sprintf) that checks for and protects
  //  against printinf lines that are too long (bigger than 4096 bytes).
  //  But some compilers may not provide this function; uncomment the following
  //  if you want to use normal printf (will crash if lines sent are > 4096 bytes)
  //  (see below for related #defines):
  //#define JrDebugDIRECTIVE_DontTryToUseSnprintf
 #endif
#endif

#ifndef JrDebugDIRECTIVE_ReplaceFunc
 #ifndef JrDebugDIRECTIVE_DontReplaceFunc
  // If your compiler does not support the __func__ macro, you need to
  //  define a replacement or uncomment this to make it blank.
  // #define JrDebugDIRECTIVE_ReplaceFunc
 #endif
#endif

#ifdef JrDebugDIRECTIVE_ReplaceFunc
 // If your compiler does not support the __func__ macro, you need to
 //  define a replacement or uncomment this to make it blank.
 #define __func__ ""
#endif
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// does the c++ compiler use snprintf or _snprintf or neither?
//  (supports the following macros you can define: JrDebugDIRECTIVE_DontUseSnprintfUnderscores, JrDebugDIRECTIVE_UseSnprintfUnderscores, JrDebugDIRECTIVE_DontTryToUseSnprintf)

// ok now map them to the underscore versions if appropriate
#ifdef JrDebugDIRECTIVE_UseSnprintfUnderscores
	#ifndef snprintf
		#define snprintf _snprintf
	#endif
	#ifndef vsnprintf
		#define vsnprintf _vsnprintf
	#endif
#endif

// or they might disable the use of them completely
#ifndef JrDebugDIRECTIVE_DontTryToUseSnprintf
	#define mysnprintf snprintf
	#define myvsnprintf vsnprintf
#else
    #define mysnprintf JrDebug::mysnprintf_rep
    #define myvsnprintf JrDebug::myvsnprintf_rep
#endif
//---------------------------------------------------------------------------


//---------------------------------------------------------------------------
// Windows switches
//
// if GetConsoleWindow() isnt available on your windows platform, comment out the next line
// 		#define JrDebugDIRECTIVE_DontUseGetConsoleWindow
//
// if you want to turn off all windows console fixing completely
//		#define JrDebugDIRECTIVE_DontFixWin32Console
//---------------------------------------------------------------------------






//---------------------------------------------------------------------------
#endif
//---------------------------------------------------------------------------
